# -*- python -*-
# $RCSfile: gtklogger.py,v $
# $Revision: 1.53 $
# $Author: reida $
# $Date: 2006/06/22 14:00:22 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

## These routines provide GUI logging and replay for PyGTK, at the gtk
## signal level.  The idea is to be able to reproduce the way in which
## a GUI drives its underlying code, without bothering to reproduce
## the way in which X11 drives the GUI.  Recording X11 events would
## produce large unreadable log files, which would break if the GUI
## were redesigned.  The philosophy here is to log only those actions
## that actually have an effect (namely, those that produce gtk
## signals that the program is connected to) and to assign those
## actions to *named* widgets.  If the GUI changes, but the names stay
## the same, there's at least a chance that a previously recorded
## script will still work properly.

## Because it's not possible (yet) to redefine pygtk class methods,
## it's necessary to make a bunch of changes to pygtk code in order to
## use the logging routines.

## All calls to GObject.connect for a signal which is emitted as a
## direct result of a user's action should be replaced by calls to
## gtklogger.connect.  Unlike GObject.connect, which returns an
## integer id, gtklogger.connect returns an object which has 'block'
## and 'unblock' methods.  These simplify the handling of signals,
## somewhat.  It is important to ensure that manipulating widgets from
## within the program does not result in logged signals.  This code

##     entry = gtk.Entry()
##     gtklogger.connect(entry, 'changed', callback)
##     entry.set_text("hello, world!")
## will create a log file entry unrelated to the user's actions.  This
## code should be used instead:
##     entry = gtk.Entry()
##     signal = gtklogger.connect(entry, 'changed', callback)
##     signal.block()
##     entry.set_text("hello, world!")
##     signal.unblock()
## (The paranoid would use a try/finally block to ensure that
## signal.unblock was called in all circumstances.)

## All gtk.Widgets that emit signals that need to be replayed have to
## have names assigned by gtklogger.setWidgetName.  (These names are
## *not* the same as those set by gtk.Widget.set_name.)  All top level
## widgets (eg, gtk.Windows) need to have names set by
## gtklogger.newTopLevelWidget.  Intermediate Widgets, such as
## gtk.Paneds and gtk.Boxes that don't emit signals but contain
## Widgets that do, can have names but aren't required to do so.  The
## only requirement is that the sequence of names, going from a top
## level widget through intermediate widgets to a signalling widget,
## uniquely specifies the signalling widget.

## GObjects that aren't Widgets but nonetheless emit signals must be
## "adopted" by an associated Widget, using gtklogger.adoptGObject.
## The arguments to adoptGObject specify how to extract the GObject
## from its adoptive parent.  The parent must be named by
## gtklogger.setWidgetName or gtklogger.newTopLevelWidget.  See the
## comments for the adoptGObject code, below, for more details.

## Instead of assigning submenus with gtk.MenuItem.set_submenu, use
## gtklogger.set_submenu(menuitem, submenu).  The submenu will behave
## as if it's been adopted by its parent menu.

## All gtk.Dialogs should be replaced by gtklogger.Dialogs.  Other
## dialogs, such as gtk.FileChooserDialog, should not be used.
## Dialogs should be named with gtklogger.newTopLevelWidget.

## Recording Log Files

## To start recording a log file, call gtklogger.start(filename).
## Call gtklogger.stop() to stop recording.
## The optional 'debugLevel' argument to start() is an integer with the
## following meanings:
##   0   No debugging output
##   1   Reports signals which can't be logged because the widget has no
##       handler, the handler doesn't handle the signal, or the widget has no
##       name.
##   2   Echoes log lines to the terminal. (Default)
##   3   Reports signals which can't be logged because the widget doesn't have
##       a top-level parent.  This is usually *not* an error.
##   4   Reports more things that are probably only useful when debugging
##       the logging code itself.

## Replaying Log Files

## To play back a log file, create a gtklogger.GUILogPlayer object and
## install it as an idle callback, like this:
##     gobject.idle_add(gtklogger.GUILogPlayer(filename))
## and then start the gtk main loop, if it's not already running.
## Optional arguments to the GUILogPlayer constructor are:
##   finishCB:   callback function to be called when done replaying
##   debugLevel: see comments for start(), above.
##   threaded:   must be set to True if the program uses threads
##   exceptHook: function to be called when an exception is raised
##               while replaying.  The arguments to the function are
##               the exception and the current GUILogLineRunner
##               instance.  The function should return True if it has
##               handled the exception, and False if the exception
##               should be propagated further.  In either case, no
##               more lines will be executed from the log file.  Note
##               that the function will be called from within a
##               gtk.threads_enter(), gtk.threads_leave() pair if
##               threaded==True.

## Instrumenting Log Files

## Sometimes it's sufficient to see if a log file just plays to
## completion without raising an exception, but often it's necessary
## to run additional tests while replaying a file.  Log files are
## passed line by line to the Python interpreter, so it's possible to
## insert single lines of Python code into them.  The log file lines
## are executed using the gtklogger module's globals dictionary, so
## anything defined in this file is accessible.  If a log file imports
## another module, it will be imported into the gtklogger namespace.

## Long lines in the log file may be continued on the next line by
## ending the first line with a backslash.  This is necessary even
## within parentheses, because the logfile parser is not a full Python
## interpreter.  Because the parser reads only one line at a time,
## it's not possible to include anything that requires indented blocks
## of Python code.  Such code should instead be moved to a function in
## a module that's imported by the log file.

## Checkpoints

## Log file lines are executed sequentially by idle callbacks in the
## gtk main loop.  If a line has side effects that occur via other
## idle callbacks, or are executed in a separate thread, it's possible
## that the *next* line of the log file may be executed before all of
## the side effects have occured.  Sometimes it's important to wait
## for the side effects.  For example, if one line of the log file
## simulates a button press, which initiates a long calculation that
## fills an array *on a separate thread*, and the next line of the log
## file performs a calculation on that array, the playback mechanism
## must know that it can't run the second line until the thread has
## finished.  It's not sufficient simply to wait for the first line to
## return, because it will return long before the subthread finishes.

## In such cases, put a call to gtklogger.checkpoint() in your code at
## the point where the long calculation *finishes*.  The argument to
## checkpoint() is a comment string of your choosing.  When recording
## a log file, the call to checkpoint() inserts a line "checkpoint
## <comment>" in the file.  When replaying, this line is not passed to
## the Python interpreter, but instead instructs the replay mechanism
## to wait until the checkpoint() function is called, with the same
## comment string.  In effect, the replay script pauses until the
## program catches up to it.

## Note that the same checkpoint comment can occur more than once.
## The script will not proceed past the nth occurrence of a checkpoint
## line unless there have been at least n corresponding calls to the
## checkpoint() function.  Also note that it's ok if a replaying
## script reaches checkpoints in a different order than they were
## reached while recording.

## The function checkpoint_count(), below, can be useful when
## debugging problems with checkpoints, especially if you're trying to
## add checkpoint lines to a script that was recorded before you
## realized that you needed checkpoint() calls in the program.

import gtk
import gobject
import os
import re
import string
import sys
import types
import weakref

_logfile = None
_replaying = False
_debugLevel = 0
_threaded = True
_suppress_motion_events = True
_old_suppression_state = True

def start(filename, debugLevel=2, suppress_motion_events=True):
    global _logfile, _debugLevel, _suppress_motion_events
    _debugLevel = debugLevel
    _suppress_motion_events = suppress_motion_events
    if _logfile is not None:
        _logfile.close()
    try:
        _logfile = open(filename, "w")
    except:
        _logfile = None
        raise

def stop():
    global _logfile
    if _logfile is not None:
        try:
            _logfile.close()
        finally:
            _logfile = None

# Utility functions to temporarily turn on motion-event recording.

def allow_motion_events():
    global _old_suppression_state
    global _suppress_motion_events
    _old_suppression_state = _suppress_motion_events
    _suppress_motion_events = False

def restore_motion_events():
    global _old_supression_state
    global _suppress_motion_events
    _suppress_motion_events = _old_suppression_state

##################################

# Dictionary of top level widgets keyed by name
_topwidgets = weakref.WeakValueDictionary()

# Any gtk.Widget that needs to have its signals logged and replayed
# must have a name, and enough of its widget ancestors must have names
# that the sequence of names uniquely identifies the widget.
# setWidgetName is used to assign the names.

def setWidgetName(widget, name):
    assert isinstance(widget, gtk.Widget)
    assert name is not None
    # When saved in the log files, widget names are surrounded by
    # single quotation marks, so we have to make sure to use only
    # double quotes in widget names.  Since colons are used as
    # delimiters in widget paths, they must be replaced too, when they
    # occur in widget names.
    safename = name.replace("'", '"').replace(":", ";")
    
    widget.oofname = safename
    return safename

def getWidgetName(widget):
    return getattr(widget, 'oofname', None)


# Top-level widgets (mostly gtk.Windows) must have their names
# assigned by newTopLevelWidget() instead of setWidgetName().

def newTopLevelWidget(name, widget):
    safename = setWidgetName(widget, name)
    widget.connect('map', _topmapCallback, name)
    _topwidgets[safename] = widget

def _topmapCallback(object, name):
    checkpoint("toplevel widget mapped %s" % name)

# gtk.GObjects which aren't gtk.Widgets but still need to have their
# signals logged should be registered with adoptGObject(). 'object' is
# the object being registered.  'parent' is an actual gtk.Widget, and
# 'access_method' is a method of the parent class that retrieves the
# adopted GObject.  For example, to make the gtk.Adjustment that's
# part of a gtk.HRange loggable, call
#    adoptGObject(range.get_adjustment(), range, range.get_adjustment)
# If there's no such parent class method, then use access_function
# instead of access_method.  The function will be called with the
# parent widget as its first argument.  In either case, additional
# arguments to the function can be passed in as access_args or
# access_kwargs.

def adoptGObject(object, parent, access_method=None, access_function=None,
                 access_args=(), access_kwargs={}):
    assert not hasattr(object, 'oofparent')
    assert parent is not None
    assert access_method is not None or access_function is not None
    object.oofparent = parent
    if access_method is not None:
        object.oofparent_access_method = access_method.__name__
    else:
        object.oofparent_access_function = access_function
    object.oofparent_access_args = access_args
    object.oofparent_access_kwargs = access_kwargs

# set_submenu should be used instead of gtk.MenuItem.set_submenu.
# This is only done within gfxmenu.py.

def set_submenu(menuitem, submenu):
    menuitem.set_submenu(submenu)
    submenu.oofparent = menuitem

# To have a gtk.GObject's actions logged, use gtklogger.connect or
# gtklogger.connect_after instead of gtk.GObject.connect.  If a widget
# that otherwise would have no callback needs to be logged, use
# gtklogger.connect_passive.  All of these functions install
# _signalLogger as a callback.

def connect(object, signal, callback, *args):
    return GUISignals(
        object,
        # Because the callbacks for different signals have different
        # numbers of arguments, and because keyword arguments don't
        # work for gtk signal handlers, but _signalLogger *does* need
        # to know what signal it's handling, we *append* the signal to
        # the arguments.  That way, we can always extract it inside
        # _signalLogger.
        object.connect(signal, _signalLogger, *(args + (signal,))),
        object.connect(signal, callback, *args)
        )

def connect_after(object, signal, callback, *args):
    return GUISignals(
        object,
        object.connect_after(signal, _signalLogger, *(args + (signal,))),
        object.connect_after(signal, callback, *args)
        )

# connect_passive should be used when an action needs to be logged and
# replayed, but wouldn't otherwise have a callback.  For example,
# keypresses in a gtk.Entry often aren't connected, but still need to
# be logged.

def connect_passive(object, signal, *args):
    return GUISignals(
        object,
        object.connect(signal, _signalLogger, *(args + (signal,)))
        )

def connect_passive_after(object, signal, *args):
    return GUISignals(
        object,
        object.connect_after(signal, _signalLogger, *(args + (signal,)))
        )

# Because connect() connects *two* signals instead of one, it would be
# a pain to keep track of them both for blocking and unblocking.  So
# instead connect() returns a GUISignals object, which holds a bunch
# of signals and can block and unblock them simultaneously.

class GUISignals:
    def __init__(self, widget, *signals):
        self.signals = signals
        self.widget = weakref.ref(widget) #, self.destroy)
    def block(self):
        map(self.widget().handler_block, self.signals)
    def unblock(self):
        map(self.widget().handler_unblock, self.signals)
    def disconnect(self):
        map(self.widget().disconnect, self.signals)
        self.signals = ()
##    def destroy(self, obj):
##        print >> sys.stderr, "Destroying", obj

# We need to be able to tell if a dialog is running (see
# GUILogLineRunner, below).  We do that by using a modified Dialog
# class that keeps track of how many dialogs are open.
        
class Dialog(gtk.Dialog):
    def __init__(self, *args, **kwargs):
        super(Dialog, self).__init__(*args, **kwargs)
        connect_passive(self, 'delete-event')
        connect_passive(self, 'configure-event')
    def run(self):
        global dialog_level             # See run_level(), below
        dialog_level += 1
        try:
            result = gtk.Dialog.run(self)
        finally:
            dialog_level -= 1
        return result
    def add_button(self, name, id):
        # Just to be nice to the programmer, turn on logging for the
        # buttons automatically.  Perhaps this should be optional.
        button = super(Dialog, self).add_button(name, id)
        assert type(name) is types.StringType
        setWidgetName(button, name)
        connect_passive(button, 'clicked')
        return button

# logScrollBars should be called on any ScrolledWindow whose scroll
# bars need to be logged.  It just encapsulates the adoptGObject calls
# that would otherwise be necessary.  It returns a pair of GUISignals
# objects, one for each scroll bar.  The return value can probably be
# ignored in most cases.

def logScrollBars(window, name=None):
    assert isinstance(window, gtk.ScrolledWindow)
    if name is not None:
        setWidgetName(window, name)
    adoptGObject(window.get_hadjustment(), window,
                 access_method=window.get_hadjustment)
    adoptGObject(window.get_vadjustment(), window,
                 access_method=window.get_vadjustment)
    return (
        connect_passive(window.get_hadjustment(), 'value-changed'),
        connect_passive(window.get_vadjustment(), 'value-changed'))


## Checkpointing.  When recording, calling "checkpoint(<comment>)"
## increments the counter and writes "checkpoint <comment>" in the log
## file.  When replaying, a "checkpoint <comment>" line in the log
## file won't finish until the same checkpoint is reached by the
## executing code.  There is no constraint on the order in which
## checkpoints are reached when running in threaded mode.

checkpointdict = {}

try:
    import threading
    checkpointlock = threading.Lock()
except ImportError:
    import dummy_threading
    checkpointlock = dummy_threading.Lock()

def checkpoint(comment):
    if _logfile is not None:            # recording
        print >> _logfile, "checkpoint", comment
        if _debugLevel >= 2:
            print >> sys.stderr, "////// checkpoint", comment
    if _replaying:
        checkpointlock.acquire()
        try:
            checkpointdict[comment] += 1
        except KeyError:
            checkpointdict[comment] = 1
        checkpointlock.release()

def reset_checkpoints():
    global checkpointdict
    checkpointdict = {}

def check_checkpoint(comment):
    # Called by CheckPointLine to see if its checkpoint has been reached.
    checkpointlock.acquire()
    try:
        try:
            count = checkpointdict[comment]
        except KeyError:
            return False
        if count > 1:
            checkpointdict[comment] = count-1
        else:
            del checkpointdict[comment]
        return True
    finally:
        checkpointlock.release()


class CheckPointException(Exception):
    pass

def checkpoint_count(comment, n=0):          # for debugging
    # If checkpoints have been changed in a program, and you want to
    # retrofit a script to use the new checkpoints, you can call this
    # function from the script to make sure that the checkpoints have
    # been inserted into the script properly.  It raises an exception
    # unless there have been n more calls to checkpoint() from the
    # code than there have been 'checkpoint' lines in the script.
    count = checkpointdict.get(comment, 0)
    print "****** checkpoint_count", comment, count
    if count != n:
        raise CheckPointException(comment)
    

## Test for uniqueness by finding all named widgets and checking that
## findWidget(gtkPath(widget))==widget.

def sanity_check(widget, path=[]):
    widgetname = getWidgetName(widget)
    if widgetname:
        path = path + [widgetname]
        if findWidget(string.join(path, ':')) is not widget:
            raise ValueError("%s is not a unique widget name" % path)
    if isinstance(widget, gtk.Container):
        for child in widget.get_children():
            sanity_check(child, path)

def comprehensive_sanity_check():
    for widget in _topwidgets.values():
        sanity_check(widget)
    

########################################################

class GtkLoggerException(Exception):
    def __init__(self, *args):
        self.msg = string.join(args, ' ')
    def __str__(self):
        return self.msg

class GtkLoggerTopFailure(GtkLoggerException):
    pass

########################################################

## Functions used while replaying to find Widgets and other GObjects

def findWidget(path):
    splitpath = path.split(':')
    top = getTopLevelWidget(splitpath[0])
    if len(splitpath) == 1:
        return top
    return _recursiveFindWidget(splitpath[1:], top)

# Given a path (list of widget names) and a base (starting widget),
# _recursiveFindGObject looks for and returns the Widget (or GObject)
# named by the path.  The search is a depth first search of the widget
# tree.
def _recursiveFindWidget(path, base):
    if not path:
        return base
    if isinstance(base, gtk.Container):
        for child in base.get_children():
            childname = getWidgetName(child)
            if childname == path[0]:
                return _recursiveFindWidget(path[1:], child)
            if childname is None:
                result = _recursiveFindWidget(path, child)
                if result is not None:
                    return result
                # Didn't find anything using unnamed widgets.  Keep
                # looking at other children.

# findAllWidgets returns the paths to all named widgets under (and
# including) the given widget.  It *won't* find adopted non-widget
# objects.  It's mostly for debugging.
def findAllWidgets(topname):
    return _findAllWidgets(findWidget(topname))
def _findAllWidgets(top):
    topname = getWidgetName(top)
    childpaths = []
    if isinstance(top, gtk.Container):
        for child in top.get_children():
            childpaths.extend(_findAllWidgets(child))
    if topname:
        childpaths = [topname] + \
                     [string.join([topname, path], ':') for path in childpaths]
    return childpaths

def getTopLevelWidget(name):
    try:
        return _topwidgets[name]
    except KeyError:
        raise GtkLoggerTopFailure(name)

def isTopLevelWidget(object):
    return object in _topwidgets.values()

def getTopWidgetNames():
    return _topwidgets.keys()[:]

def findMenu(widget, path):
    # widget is a gtk.MenuShell (container for MenuItems)
    splitpath = path.split(':', 1)
    for child in widget.get_children(): # child is a gtk.MenuItem
        if getWidgetName(child) == splitpath[0]:
            if len(splitpath) == 1:
                return child
            return findMenu(child.get_submenu(), splitpath[1])

def findCellRenderer(treeview, col, rend):
    return treeview.get_column(col).get_cell_renderers()[rend]

# Utility function for creating a gtk.gdk.Event object.  "etype" must
# be an event type (gtk.gdk.BUTTON_PRESS, for example).  "kwargs" can
# contain attributes of the event object.  It almost certainly should
# include the "window" attribute, which must be set to a
# gtk.gdk.Window.  (For gtk.Widgets, this is just Widget.window.  For
# adopted GObjects, it's harder to get the correct gdk.Window into the
# log...)

def event(etype, **kwargs):
    ev = gtk.gdk.Event(etype)
    for arg, val in kwargs.items():
        setattr(ev, arg, val)
    return ev

#######################################################################

_loggerDir = []

# _loggerDir is a sorted list of (gtk class, GtkLogger class)
# pairs.  The list is sorted so that each gtk class comes before its
# base classes, so that the GtkLogger corresponding to the most
# derived class is found first.

def findLogger(object):
    for klass, logger in _loggerDir:
        if isinstance(object, klass):
            return logger()
    raise NotImplementedError("No GtkLogger for %s" % object.__class__.__name__)

# localvar generates names for local variables in scripts.
_localvarcount = {}

def localvar(base):
    count = _localvarcount.get(base, 0)
    _localvarcount[base] = count + 1
    return "%s_%d" % (base, count)
        
# GtkLoggerMetaClass is the metaclass for GtkLogger.  It ensures that
# all GtkLogger classes are listed in the _loggerDir in the right
# order.  It requires that each GtkLogger subclass have a class-level
# tuple called 'classes' which contains the gtk classes to which the
# Logger applies.

class GtkLoggerMetaClass(type): 
    def __init__(cls, name, bases, dict):
        super(GtkLoggerMetaClass, cls).__init__(name, bases, dict)
        # get the list of gtk classes to which this logger applies
        klasses = dict.get('classes', ())
        # Make a _loggerDir entry for each gtk class.
        for klass in klasses:
            # Insertion sort!  The sorting criterion (subclasses
            # before base classes) doesn't define a unique ordering,
            # and we don't care about speed here, anyway.
            for i in range(len(_loggerDir)):
                if issubclass(klass, _loggerDir[i][0]):
                    _loggerDir.insert(i, (klass, cls))
                    break
            else:
                _loggerDir.append((klass, cls))
                

class GtkLogger(object):
    __metaclass__ = GtkLoggerMetaClass
    def location(self, object, *args):
        ## location() returns a string that can be evaluated by Python
        ## to return the given object.  The string specifies the
        ## object's location, in some sense.
        raise GtkLoggerException("No location function")

    def record(self, object, signal, *args):
        ## record() returns a *list* of strings that can be evaluated
        ## to recreate a gui action.  Each string is a single Python
        ## line.  Only Python statements that can be evaluated in a
        ## single line are allowed, because they aren't being read by
        ## a full Python parser.  That means that
        ##    tree = findWidget('window:tree')
        ##    tree.row_activated(...)
        ## is ok, but
        ##    if tree is not None:
        ##       tree.row_activated(...)
        ## is not.   Import statements and function calls are allowed.
        raise GtkLoggerException("No record function for",
                                 object.__class__.__name__, signal,
                                 "in", self.__class__.__name__)
    # GtkLogger.record() should return GtkLogger.ignore if an event
    # should not be logged.
    class Ignore:
        pass
    ignore = Ignore()

class WidgetLogger(GtkLogger):
    classes = (gtk.Widget,)
    events = {"button-press-event"  : "BUTTON_PRESS",
              "button-release-event": "BUTTON_RELEASE",
              "motion-notify-event" : "MOTION_NOTIFY"}
    buttonup = True
    def location(self, widget, *args):
        name = getWidgetName(widget)
        if not name:
            raise GtkLoggerException("Unnamed widget")
        path = self._parentWidgetPath(widget) + [name]
        if path[0] not in getTopWidgetNames():
            raise GtkLoggerException(string.join(path, ':') + 
                                     " is not contained in a top-level widget")
        return "findWidget('%s')" % string.join(path, ':')
    def _parentWidgetPath(self, widget):
        parent = widget.get_parent()
        while parent is not None and getWidgetName(parent) is None:
            parent = parent.get_parent()
        if parent is None:
            return []
        return self._parentWidgetPath(parent) + [getWidgetName(parent)]
    def record(self, object, signal, *args):
        eventname = self.events.get(signal, None)
        if eventname is not None:
            evnt = args[0]
            if eventname == 'BUTTON_PRESS':
                WidgetLogger.buttonup = False
            elif eventname == 'BUTTON_RELEASE':
                WidgetLogger.buttonup = True
            elif eventname == 'MOTION_NOTIFY':
                if suppress_motion_events and WidgetLogger.buttonup:
                    return self.ignore
            wvar = localvar('widget')
            return [
                "%s = %s" % (wvar, self.location(object, *args)),
                "%s.event(event(gtk.gdk.%s,x=%20.13e,y=%20.13e,state=%d,window=%s.window))" % (wvar, eventname, evnt.x, evnt.y, evnt.state, wvar)
##                "%s.event(event(gtk.gdk.%s,x=%20.13e,y=%20.13e,state=%d))"
##                    %(self.location(object, *args), eventname,
##                      evnt.x, evnt.y, evnt.state)
                ]

        if signal == 'focus_in_event':
            wvar = localvar('widget')
            return [
       "%s=%s" % (wvar, self.location(object, *args)),
       "%(widget)s.event(event(gtk.gdk.FOCUS_CHANGE, in_=1, window=%(widget)s.window))" % dict(widget=wvar)
                ]
        if signal == 'focus_out_event':
            wvar = localvar('widget')
            return [
       "%s=%s" % (wvar,self.location(object, *args)),
       "%(widget)s.event(event(gtk.gdk.FOCUS_CHANGE, in_=0, window=%(widget)s.window))" % dict(widget=wvar)
                ]
            
        # To log the changes in a Paned's divider position, connect to
        # the 'size-allocate' signal of one of its children.  The
        # Paned will need a name.  The child doesn't.
        
        if signal == 'size-allocate':
            alloc = object.get_allocation()
            parent = object.get_parent()
            if isinstance(parent, gtk.Paned):
                parentlogger = findLogger(parent)
                return ["%s.set_position(%d)" % (parentlogger.location(parent),
                                                parent.get_position())]
            return ["%s.size_allocate(gtk.gdk.Rectangle(%d, %d, %d, %d))" \
                   % (self.location(object, *args),
                      alloc.x, alloc.y, alloc.width, alloc.height)]
        return super(WidgetLogger, self).record(object, signal, *args)

class WindowLogger(WidgetLogger):
    classes = (gtk.Window,)
    def record(self, object, signal, *args):
        # Don't log "destroy" signals, because they're sent for every
        # object that's destroyed, not just the one that initiated the
        # action.
        if signal == 'destroy':
            return self.ignore
        if signal == 'delete-event':

            # Simply re-emitting the delete event doesn't actually
            # destroy a window, so we have to explicitly destroy it.
            # This may be the wrong thing to do if a window intercepts
            # delete-event and doesn't propagate the signal.
            
##            return ["%s.destroy()" % self.location(object, *args)]
            wvar = localvar('widget')
            hvar = localvar('handled')
            return [
                "%s=%s" % (wvar, self.location(object, *args)),
                "%(h)s=%(w)s.event(event(gtk.gdk.DELETE,window=%(w)s.window))"
                % dict(w=wvar, h=hvar),
                "postpone if not %(h)s: %(w)s.destroy()" % dict(w=wvar, h=hvar),
                    ]
        if signal == 'configure-event':
            event = args[0]
            return ["%s.resize(%d, %d)" % (self.location(object, *args),
                                          event.width, event.height)]
        return super(WindowLogger, self).record(object, signal, *args)

class ButtonLogger(WidgetLogger):
    classes = (gtk.Button,)
    def record(self, object, signal, *args):
        if signal == 'clicked':
            return ["%s.clicked()" % self.location(object, *args)]
        return super(ButtonLogger, self).record(object, signal, *args)

class RadioButtonLogger(WidgetLogger):
    classes = (gtk.RadioButton,)
    def record(self, object, signal, *args):
        if signal == 'clicked':
            # When a radio button is pressed, two 'clicked' signals
            # are sent, one to deactivate the old button and one to
            # activate the new one.  We only want to log the new one.
            if object.get_active():
                return ["%s.clicked()" % self.location(object, *args)]
            return self.ignore
        return super(RadioButtonLogger, self).record(object, signal, *args)
        
class EntryLogger(WidgetLogger):
    classes = (gtk.Entry,)
    def record(self, object, signal, *args):
        if signal == 'changed':
            text = object.get_text().replace('\\', '\\\\')
            return ["%s.set_text('%s')" % (self.location(object, *args),text)]
        return super(EntryLogger, self).record(object, signal, *args)

class ComboBoxLogger(WidgetLogger):
    classes = (gtk.ComboBox,)
    def record(self, object, signal, *args):
        if signal == 'changed':
            index = object.get_active()
            return ["%s.set_active(%d)" % (self.location(object, *args), index)]
        return super(ComboBoxLogger, self).record(object, signal, *args)

class ExpanderLogger(WidgetLogger):
    classes = (gtk.Expander,)
    def record(self, object, signal, *args):
        if signal == 'activate':
            return ["%s.set_expanded(%d)" % (self.location(object, *args),
                                            object.get_expanded())]
        return super(ExpanderLogger, self).record(object, signal, *args)

class FileChooserLogger(WidgetLogger):
    classes = (gtk.FileChooserWidget, gtk.FileChooserDialog)
    # Although FileChooserDialog is listed here, it's not really
    # loggable because it doesn't use gtklogger.Dialog as a base
    # class.
    def record(self, object, signal, *args):
        if signal == "selection-changed":
            filename = object.get_filename() or ""
            return ["%s.set_filename('%s')" % (self.location(object, *args),
                                               filename)]
        return super(FileChooserLogger, self).record(object, signal, *args)

class MenuLogger(GtkLogger):
    classes = (gtk.MenuItem,)
    def location(self, menuitem, *args):
        parent, path = self._getMenuPath(menuitem)
        parentcode = findLogger(parent).location(parent)
        return "findMenu(%s, '%s')" % (parentcode, string.join(path, ':'))
    def record(self, object, signal, *args):
        if signal == "activate":
            return ["%s.activate()" % self.location(object, args)]
        return super(MenuLogger, self).record(object, signal, *args)
    
    # Find a list of menu item names leading to the given
    # gtk.MenuItem.  This relies on setting gtk.Menu.oofparent in
    # gtklogger.set_submenu.  The return value is a tuple containing
    # the parent widget of the top of the menu hierarchy and the list
    # of menuitem names.
    def _getMenuPath(self, gtkmenuitem):
        path = [getWidgetName(gtkmenuitem)]
        parent = gtkmenuitem.get_parent()
        if isinstance(parent, gtk.Menu) and \
               isinstance(parent.oofparent, gtk.MenuItem):
            pparent, ppath = self._getMenuPath(parent.oofparent)
            return pparent, ppath+path
        else:
            return parent, path

class AdopteeLogger(GtkLogger):
    #  Handles non-Widgets that were "adopted" by adoptGObject().
    def location(self, object, *args):
        parent = getattr(object, 'oofparent')
        parentcode = findLogger(parent).location(parent)
        strargs = [`x` for x in object.oofparent_access_args] + \
                  ["%s=%s" % (name, `val`)
                   for name, val in object.oofparent_access_kwargs.items()]
        if hasattr(object, 'oofparent_access_method'):
            return '%s.%s(%s)' % (parentcode, object.oofparent_access_method,
                                  string.join(strargs, ','))
        else: # must have oofparent_access_function instead
            return '%s(%s)' % (object.oofparent_access_function,
                               string.join([parentcode]+strargs,', ')) 

class AdjustmentLogger(AdopteeLogger):
    classes = (gtk.Adjustment,)
    def record(self, object, signal, *args):
        if signal == 'value-changed':
            return ['%s.set_value(%20.13e)' % (self.location(object, *args),
                                               object.get_value())]

        return super(AdjustmentLogger, self).record(object, signal, *args)

class TreeViewLogger(WidgetLogger): # Oh, I'm a lumberjack and I'm OK...
    classes = (gtk.TreeView,)
    def record(self, object, signal, *args):
        if signal == 'button-release-event':
            event = args[0]
            wvar = localvar('widget')
            return [
    "%s=%s" % (wvar, self.location(object, *args)),
    "%(w)s.event(event(gtk.gdk.BUTTON_RELEASE,button=%(b)d,window=%(w)s.window))"
    % dict(w=wvar, b=event.button)
    ]
        if signal == 'row-activated':
            path = args[0]
            col = args[1]               # gtk.TreeViewColumn obj
            cols = object.get_columns()
            for i in range(len(cols)):
                # TODO is there a better way to find the column number?
                if cols[i] is col:
                    return ["tree=%s" % self.location(object, *args),
                            "column = tree.get_column(%d)" % i,
                            "tree.row_activated(%s, column)" % `path`]
        if signal == 'row-expanded':
            path = args[1]
            return ["%s.expand_row(%s, open_all=False)"
                    % (self.location(object, *args), `path`)]
        if signal == 'row-collapsed':
            path = args[1]
            return ["%s.collapse_row(%s)"
                    % (self.location(object,*args), `path`)]
        return super(TreeViewLogger, self).record(object, signal, *args)

class TreeViewColumnLogger(AdopteeLogger):
    classes = (gtk.TreeViewColumn,)

class TreeSelectionLogger(AdopteeLogger):
    classes = (gtk.TreeSelection,)
    def record(self, object, signal, *args):
        if signal == 'changed':
            if object.get_mode()==gtk.SELECTION_MULTIPLE:
                model, rows = object.get_selected_rows()
                # Unselecting all rows and then selecting the selected
                # ones seems wrong, since unselecting and reselecting
                # might have side effects.  It would be much better
                # just to select or unselect the changed rows, but I
                # don't see how to do that simply.
                return ["%s.unselect_all()" % self.location(object, *args)] + \
                       ["%s.select_path(%s)" % (self.location(object,*args),row)
                        for row in rows]
            else:                       # single selection only
                model, iter = object.get_selected()
                if iter is not None:
                    path = model.get_path(iter)
                    return ["%s.select_path(%s)" 
                            % (self.location(object, *args), path)]
                else:
                    return ["%s.unselect_all()" % self.location(object, *args)]
        return super(TreeSelectionLogger, self).record(object, signal, *args)

class ListStoreLogger(AdopteeLogger):
    classes = (gtk.ListStore,)
    insertrow = None                      # destination for row drag'n'drop
    def record(self, object, signal, *args):
        # Drag-and-drop of a line within a ListStore creates a pair of
        # row-inserted and row-deleted signals, with row-inserted
        # coming first.  The two must be logged together, so we don't
        # actually return a non-trivial result until getting the
        # second signal.  If the user has two mice and drags rows of
        # two ListStores simultaneously, this won't work.
        if signal == "row-inserted":
            ListStoreLogger.insertrow = args[0][0] # args[0] is a gtk tree path
            return self.ignore
        if signal == "row-deleted":
            if ListStoreLogger.insertrow is None:
                return self.ignore
            deleterow = args[0][0]      # args[0] is a gtk tree path
            destrow = ListStoreLogger.insertrow # destination row
            # The row to be deleted contains the data that has to be
            # inserted in the new row. "deleterow" was set after the
            # new row was inserted (because 'row-deleted' was sent
            # after 'row-inserted'), so if the new row comes before
            # the deleted row in the list, then on replay, where
            # nothing has happened yet, the row to be deleted has a
            # different index than the source row.
            if ListStoreLogger.insertrow < deleterow:
                sourcerow = deleterow - 1
            else:
                sourcerow = deleterow
            ListStoreLogger.insertrow = None
            lvar = localvar('ls')
            dvar = localvar('data')
            return [
"%s = %s" % (lvar, self.location(object, *args)),
"%(data)s = [%(ls)s.get_value(%(ls)s.get_iter((%(r)d,)),i) for i in range(%(ls)s.get_n_columns())]" % dict(r=sourcerow, data=dvar, ls=lvar),
"%s.insert(%d, %s)" % (lvar, destrow, dvar),
"%(ls)s.remove(%(ls)s.get_iter((%(r)d,)))" % dict(r=deleterow, ls=lvar)
                ]

        return super(ListStoreLogger, self).record(object, signal, *args)

class CellRendererLogger(AdopteeLogger):
    classes = (gtk.CellRenderer,)
    def record(self, object, signal, *args):
        if signal == 'toggled':
            path = args[0]
            return ["%s.emit('toggled', %s)"
                    % (self.location(object, *args), `path`)]
        return super(CellRendererLogger, self).record(object, signal, *args)


##############################################################

##############################################################
##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##
##############################################################    

# This is the gtk signal handler that records everything.

def _signalLogger(object, *args):
##    print "_signalLogger:", object, args
    if _logfile:
        # The signal was appended to the arg list by connect().
        signal = args[-1]
        try:
            records = findLogger(object).record(object, signal, *args[:-1])
        except GtkLoggerTopFailure, exc:
            if _debugLevel >= 3:
                print >> sys.stderr, "Can't log %s (%s): %s" \
                      (object.__class__.__name__, signal, exc)
        except GtkLoggerException, exc:
            if _debugLevel >= 1:
                print >> sys.stderr, "Can't log %s (%s): %s" % \
                      (object.__class__.__name__, signal, exc)
        else:
            if records is GtkLogger.ignore:
                pass
            elif records is not None:
                assert type(records) is types.ListType
                for record in records:
                    _writeline(record)
            else:
                if _debugLevel >= 1:
                    print >> sys.stderr, "No record function for", object, signal
    return False                        # propagate events


## Some lines are logged too often.  This code eliminates redundant
## lines, making the log files easier to read and faster to run.

## _redundantlines is a list of regular expression objects that match
## lines that are redundant when repeated.
_redundantlines = [
    # Each time a Window is created, it emits four 'configure-event'
    # signals, which show up as four 'resize' lines in the log.
    re.compile(r".*\.resize\([0-9]+, [0-9]+\)$"),
    # set_position lines for HPaned and VPaned widgets are generated
    # from their children's 'size-allocate' signals, which are
    # repeated unnecessarily, as far as we're concerned.
    re.compile(r".*\.set_position\([0-9]+\)$")
    ]

def _not_redundant(line):
    for regexp in _redundantlines:
        if regexp.match(line) is not None:
            return False
    return True

_prevline = None
def _writeline(line):
    global _prevline
    if line != _prevline or _not_redundant(line):
        print >> _logfile, line
        if _debugLevel >= 2:
            print >> sys.stderr, "//////", line
    _prevline = line

##############################################################
##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##  ##
##############################################################    

# replaydelay is the delay time (in milliseconds) that is inserted
# between lines in the log file during playback, if there is no
# explicit delay ('pause' statement).  The timing is not guaranteed to
# be precise.
replaydelay = 0

def set_delay(delay):
    global replaydelay
    replaydelay = delay

# Time in milliseconds to wait before retrying a checkpoint or looking
# for a window.
retrydelay = 100

# maxtries limits the number of times a line that raises a
# GtkLoggerTopFailure exception will be retried.
maxtries = 100

# Keep track of the run level, which is the gtk run level plus the
# number of open modal dialogs.  See the comment in
# GUILogLineRunner.__call__.
dialog_level = 0
def run_level():
    global dialog_level
    return gtk.main_level() + dialog_level

# A GUILogPlayer reads a log file of saved gui events and simulates them.
# It should be installed as an idle callback, like this:
#      gobject.idle_add(GUILogPlayer(logfilename))
# Optional arguments to the constructor are:
#   finishCB:   callback function to be called when done replaying
#   debugLevel: see comments for start(), above.
#   threaded:   must be set to True if the program uses threads
#   exceptHook: see comments above.

class GUILogPlayer(object):
    current = None
    def __init__(self, filename, finishCB=None, debugLevel=2, threaded=False,
                 exceptHook=None):
        global _debugLevel, _threaded
        _debugLevel = debugLevel
        _threaded = threaded
        self.finishCB = finishCB
        self.exceptHook = exceptHook
        try:
            file = open(filename, "r")
        except IOError:
            self.aborted = True
            raise
        self.linerunners = []

        # More than one execution line can come from a single source
        # line, if, for example, automatic pauses are inserted.  The
        # execution line number is used to know when to run a line.
        # The source line number is useful for debugging.
        lineno = 0                      # number of the line being executed
        srcline = 1                     # source line number
        pausenext = False               # add a pause before the next line?
        lines = file.readlines()
        self._nfilelines = len(lines)
        tobecontinued = ""
        for line in lines:
            line = line.rstrip()
            if line and line.lstrip()[0] != "#":
                if line[-1] == '\\':
                    # Lines ending in a backslash aren't processed
                    # immediately.  They're prepended to the next line
                    # instead.
                    tobecontinued += line[:-1]
                else:
                    line = tobecontinued + line
                    tobecontinued = ""
                    try:
                        words = line.split(None, 1)
                    except:
                        words = None
                    if words and words[0] == 'pause':
                        self.linerunners.append(
                            PauseLine(self, srcline, lineno, eval(words[1])))
                        lineno += 1
                        pausenext = False
                    elif words and words[0] == "checkpoint":
                        self.linerunners.append(
                            CheckPointLine(self, srcline, lineno,
                                           words[1].rstrip()))
                        lineno += 1
                        pausenext = False
                    elif words and words[0] == "postpone":
                        self.linerunners.append(
                            PostponeLine(self, srcline, lineno, words[1]))
                        lineno += 1
                    else:
                        if pausenext and replaydelay > 0:
                            self.linerunners.append(
                                PauseLine(self, srcline, lineno, replaydelay))
                            lineno += 1
                        self.linerunners.append(
                             PerformLine(self, srcline, lineno, line.rstrip()))
                        lineno += 1
                        pausenext = True
            srcline += 1
        file.close()
        GUILogPlayer.current = self
        self.aborted = False
    def __del__(self):
        GUILogPlayer.current = None
    def __call__(self):
        global _replaying
        _replaying = True
        if len(self.linerunners) > 0:
            self.linerunners[0].start()
        return False                    # only call this callback once
    def stop(self):
        global _replaying
        if _replaying:
            _replaying = False
            if self.finishCB:
                self.finishCB()
    def abort(self):
        self.aborted = True
        self.stop()
    def getLine(self, n):
        if n >= 0 and n < len(self.linerunners):
            return self.linerunners[n]
    def nlines(self):
        return self._nfilelines

# A GUILogLineRunner is in charge of executing a single line of the
# gui log file.

class GUILogLineRunner(object):
    def __init__(self, logrunner, srcline, lineno):
        self.lineno = lineno
        self.srcline = srcline
        self.logrunner = logrunner
        self.ntries = 0
        self.status = "initialized"
        # insertedLines contains GUILogLineRunners for postponed lines
        # that have been run just before this line.
        self.insertedLines = []
    def start(self):
        # Install ourself as an idle callback.  The callback is set to
        # a low priority, because we're simulating a user's mouse
        # clicks and keyboard events, and users are slow.  If we run
        # at normal priority, the simulated events interfere with
        # normal gtk operation.
        if self.status == "initialized":
            self.status = "installed"
            if _debugLevel >= 4:
                print >> sys.stderr, "Installing", self.srcline
            gobject.idle_add(self, priority=gobject.PRIORITY_LOW)
##    def prevLine(self):
##        if not self.insertedLines:
##            return self.logrunner.getLine(self.lineno-1)
##        return self.insertedLines[-1]
    def nextLine(self):
        line = self.logrunner.getLine(self.lineno+1)
        if line:
            return line
        if _postponed:
            return _postponed[-1]
    def nlines(self):
        return self.logrunner.nlines()
    def ready(self):
        if not self.insertedLines:
            pl = self.logrunner.getLine(self.lineno-1)
            return (not pl or
                    pl.status == "done" or
                    (pl.status == "running" and pl.runLevel < run_level()))
        for line in self.insertedLines:
            if not (line.status == "done" or
                    line.status == "running" and line.runLevel < run_level()):
                return False
        return True

    def __call__(self):
        # Execute our line of the gui log, *if* the previous line has
        # completed.  Figuring out if the previous line has completed
        # is non-trivial, because the previous line may have emitted a
        # gtk signal that caused a modal dialog box to open, in which
        # case its "emit" call won't return until the box has closed!
        # *This* line contains the commands that operate the dialog
        # box, and must be issued even though the previous command
        # hasn't returned.  If the previous line hasn't returned it
        # must have called Dialog.run or started up a new gtk main
        # loop, so by keeping track of gtk.main_level() and the number
        # of open dialogs, we can tell when it's time to execute our
        # line.  (This is why we must redefine the Dialog class.)
        if self.logrunner.aborted:
            # The previous line raised an exception, so don't run this
            # line, even though it's already been installed as an idle
            # callback.
            self.status = "aborted"
            return False

        # Run this line, but only if there are no postponed lines
        # ready to go, and if this line is also ready.
        if not self.run_postponed() and self.ready():
            assert self.status in ("repeating", "installed")
            # Add the idle callback for the next line *before*
            # executing our line, because we might not return
            # until after the next line is done!  This is why we
            # need a separate idle callback for each line.
            if self.status != "repeating" and self.nextLine() is not None:
                self.nextLine().start()

            if _threaded:
                gtk.gdk.threads_enter()
            try:
                if self.status == "installed":
                    self.status = "running"
                    self.report()
                self.runLevel = run_level()
                # self.playback performs some suitable action and
                # returns True if the idle callback should be
                # repeated, and False if it shouldn't.  It can also
                # reinstall the callback, and should set self.status
                # to "done" if the task is finished.
                try:
                    result = self.playback()
                    if self.nextLine() is None:
                        self.logrunner.stop()
                    return result
                except GtkLoggerTopFailure:
                    # It's possible that the previous log line tried
                    # to open a window, but the window hasn't actually
                    # appeared yet.  In that case, our line will have
                    # failed with a GtkLoggerTopFailure exception.  We
                    # just want to keep trying (within reason) until
                    # the window appears.  Using checkpoints to wait
                    # until the window is mapped makes this problem
                    # less frequent, but doesn't make it go away
                    # entirely.
                    self.status = "repeating"
                    self.ntries += 1
                    if self.ntries == maxtries:
                        if _debugLevel >= 1:
                            print >> sys.stderr, \
                                  "Failed to find top-level widget after", \
                                  self.ntries, "attempts."
                        raise
                    # Keep trying.  By reinstalling ourself in the
                    # idle callback table and returning False
                    # (meaning, "don't repeat this callback") we move
                    # to the back of the queue.  This allows the
                    # widget we are waiting for to appear, we hope.
                    gobject.timeout_add(retrydelay,
                                        self, priority=gobject.PRIORITY_LOW)
                    return False             
                except Exception, exc:
                    # Any type of exception other than GtkLoggerTopFailure
                    # is fatal.
                    self.status = "aborted"
                    self.logrunner.abort()
                    if self.logrunner.exceptHook:
                        if not self.logrunner.exceptHook(exc, self):
                            raise exc
            finally:
                gtk.gdk.flush()
                if _threaded:
                    gtk.gdk.threads_leave()
        # We're still waiting for the previous line to execute. We put
        # ourself at the back of the execution queue (by reinstalling
        # and returning False) so that the previous line will run
        # first.
        if _debugLevel >= 4:
            print >> sys.stderr, "Reinstalling", self.srcline
        gobject.timeout_add(retrydelay, self, priority=gobject.PRIORITY_LOW)
        return False

    def run_postponed(self):
        ## Attempt to run a postponed line.  Return True if successful.
        if not _postponed:
            return False                # no postponed lines to run
        line = _postponed[-1]
        if line.ready():
            # Put the postponed line in the list of lines that must
            # run before this line is run.
            self.insertedLines.append(line)
            # Start the idle callback that will run the postponed line.
            line.start()
            return True
        return False                    # postponed line isn't ready

class PerformLine(GUILogLineRunner):
    def __init__(self, logrunner, srcline, lineno, line):
        GUILogLineRunner.__init__(self, logrunner, srcline, lineno)
        self.line = line
    def report(self):
        print >> sys.stderr, "////// %d/%d %s" %(self.srcline, self.nlines(),
                                                 self.line)
    def playback(self):
        # Exec'ing the line with an explicitly provided dictionary
        # allows variables created on one line to be available on a
        # later line.  Otherwise, the variable's scope would just be
        # this function call, which wouldn't be very useful.
        if _debugLevel >= 4:
            print >> sys.stderr, "Executing", self.srcline, self.line
        exec(self.line, sys.modules[__name__].__dict__)
        if _debugLevel >= 4:
            print >> sys.stderr, "Finished", self.srcline, self.line
        self.status = "done"
        return False

_postponed = []

class PostponeLine(GUILogLineRunner):
    def __init__(self, logrunner, srcline, lineno, line):
        GUILogLineRunner.__init__(self, logrunner, srcline, lineno)
        self.line = line
        self.runlevel = run_level()
    def playback(self):
        _postponed.append(PostponedLine(self))
        self.status = "done"
        return False
    def report(self):
        print >> sys.stderr, "////// %d/%d postponing %s" % (self.srcline,
                                                             self.nlines(),
                                                             self.line)

class PostponedLine(PerformLine):
    def __init__(self, ppl):
        # ppl is a PostponeLine object
        PerformLine.__init__(self, ppl.logrunner, ppl.srcline, ppl.lineno,
                             ppl.line)
        self.ppl = ppl
    def ready(self):
        # Overrides PerformLine.ready()
        return run_level() <= self.ppl.runlevel
    def start(self):
        _postponed.remove(self)
        PerformLine.start(self)
    def playback(self):
        PerformLine.playback(self)
    def run_postponed(self):
        # Postponed lines never wait for other postponed lines.
        return False
    def report(self):
        print >> sys.stderr, "////// %d/%d (postponed) %s" % (self.srcline,
                                                            self.nlines(),
                                                            self.line)

class PauseLine(GUILogLineRunner):
    # Special handler for lines of the form "pause <time>".  Such
    # lines allow threads started by earlier commands to complete.
    # The pause is guaranteed to be at least <time> milliseconds long.
    def __init__(self, logrunner, srcline, lineno, delaytime):
        GUILogLineRunner.__init__(self, logrunner, srcline, lineno)
        self.delaytime = delaytime
    def playback(self):
        if _threaded and self.delaytime > 0:
            if self.status == "running":
                self.status = "repeating"
                if _debugLevel >= 4:
                    print >> sys.stderr, self.srcline, \
                          "Pausing", self.delaytime, "milliseconds"
                gobject.timeout_add(self.delaytime, self,
                                    priority=gobject.PRIORITY_LOW)
            elif self.status == "repeating":
                if _debugLevel >= 4:
                    print >> sys.stderr, "Done pausing", self.srcline
                self.status = "done"
        else:
            # not threaded, no need to wait for background tasks
            self.status = "done"
        return False
    def report(self):
        pass

class CheckPointLine(GUILogLineRunner):
    def __init__(self, logrunner, srcline, lineno, comment):
        GUILogLineRunner.__init__(self, logrunner, srcline, lineno)
        self.comment = comment
    def playback(self):
        if check_checkpoint(self.comment):
            if _debugLevel >= 4:
                print >> sys.stderr, "Reached checkpoint", self.srcline
            self.status = "done"
        else:
            self.status = "repeating"
            if _debugLevel >= 4:
                print >> sys.stderr, "Waiting on checkpoint", self.srcline
            gobject.timeout_add(retrydelay, self,
                                priority=gobject.PRIORITY_LOW)
        return False
    def report(self):
        print >> sys.stderr, "////// %d/%d checkpoint %s" %(self.srcline,
                                                            self.nlines(),
                                                            self.comment)
