# -*- python -*-
# $RCSfile: setup.py,v $
# $Revision: 1.63 $
# $Author: langer $
# $Date: 2006/07/05 17:09:28 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

## Usage:

#  In the top oof2 directory (the one containing this file) type
#  something like this:
#    python setup.py install     # installs in the default location
#    python setup.py install --prefix=/some/other/place
#    python setup.py [build [--debug]] install --prefix ...
#  The flags --enable-mpi, --enable-petsc, and --enable-devel can
#  occur anywhere after 'setup.py' in the command line.


# External utilities
MAKEDEPEND = 'makedepend'

# The following determine some secondary installation directories.
# They will be created within the main installation directory
# specified by --prefix. 

DATADIR = "share/oof2"
EXAMPLEDIR = "share/oof2/examples"
DOCDIR = "share/oof2/doc"

# Required version numbers of required external libraries.  These
# aren't used explicitly in this file, but they are used in the DIR.py
# files that are execfile'd here.

GTK_VERSION = "2.6.0"
PYGTK_VERSION = "2.6"
GNOMECANVAS_VERSION = "2.6"

###############################

import distutils.core
from distutils.command import build
from distutils.command import build_ext
from distutils.command import build_py
from distutils.command import clean
from distutils import errors
from distutils import log
from distutils.dir_util import remove_tree
from distutils.sysconfig import get_config_var

import shlib # adds build_shlib and install_shlib to the disutils command set
from shlib import build_shlib

from oof2setuputils import run_swig, find_file

import os
import stat
import string
import sys
import tempfile

# Tell distutils that .C is a C++ file suffix.
from distutils.ccompiler import CCompiler
CCompiler.language_map['.C'] = 'c++'

# py2app stuff is commented out because, in python 2.3 at least,
# py2app installs itself in a way that's inconsistent with
# build_shlib, and build_shlib is more important for us.
##try:
##    import py2app                       # for creating a frozen Mac app.
##except ImportError:
##    have_py2app = False
##else:
##    have_py2app = True

DIRFILE = "DIR.py"                      # oof subdirectory manifest files
SWIGDIR = "SWIG"                        # root dir for swig output, inside SRC
SWIGCFILEEXT = 'cmodule.C'              # suffix for swig output files


##############

# readDIRs() walks through the oof2 source directory looking for
# DIR.py files, reads the files, and fills in the CLibInfo objects.
# CLibInfo categorize all of the source files except for the python
# files for pure python modules.

# DIR.py files contain the following python variables.  All are optional.

# dirname: The name of the directory. It's actually not used.

# clib: the name of the library to be built from the C and C++ files
# in the directory.  The library will be called lib<name>.<suffix>,
# where suffix is system dependent.  More than one DIR.py file can use
# the same name.

# cfiles: a list of the names of all of the C and C++ files in the
# directory that need to be compiled to form lib<name>.

# hfiles: a list of the names of the header files that go with the C
# and C++ files.

# swigfiles: a list of the names of the swig input files in the
# directory.  Swig-generated C++ code will be compiled but *not*
# included in lib<name>.  Each swig input file will create a separate
# python-loadable module which will *link* to lib<name>.

# swigpyfiles: a list of the names of python files that are included
# in swig output files.

# clib_order: an integer specifying the order in which the libraries
# must be built.  Later libraries may link to earlier ones.  This
# linking is done by setting clib.externalLibs in the set_clib_flags
# function in a DIR.py file.

# set_clib_flags: a function which may be called to set compilation
# and linker flags for building the library.  Its argument is a
# CLibInfo object.  The includeDirs, externalLibDirs, and externalLibs
# members of the object may be modified by set_clib_flags.

# subdirs: a list of subdirectories that should be processed.  Each
# subdirectory must have its own DIR.py file.

# TODO: ensure that typemaps.swg files are included in the source
# distribution!

allCLibs = {}
purepyfiles = []

def getCLibInfo(name):
    try:
        return allCLibs[name]
    except KeyError:
        clib = allCLibs[name] = CLibInfo(name)
        return clib
                
class CLibInfo:
    def __init__(self, name):
        allCLibs[name] = self
        self.libname = name
        self.dirdata = {'cfiles': [],   # *.[Cc] -- c and c++ source code
                        'hfiles': [],   # *.h    -- c and c++ header files
                        'swigfiles': [], # *.swg  -- swig source code
                        'swigpyfiles': [], # *.spy  -- python included in swig
                        }
        self.externalLibs = []
        self.externalLibDirs = []
        self.includeDirs = []
        self.extensionObjs = None
        self.ordering = None

    # Parse the file lists in a DIR.py file.  The file has been read
    # already, and its key,list pairs are in dirdict.  Only the data
    # relevant to CLibInfo is dealt with here.  The rest is handled
    # by readDIRs().
    def extractData(self, srcdir, dirdict):
        for key in self.dirdata.keys():
            try:
                value = dirdict[key]
                del dirdict[key]
            except KeyError:
                pass
            else:
                for filename in value:
                    self.dirdata[key].append(os.path.join(srcdir, filename))
        try:
            flagFunc = dirdict['set_clib_flags']
            del dirdict['set_clib_flags']
        except KeyError:
            pass
        else:
            flagFunc(self)
        try:
            self.ordering = dirdict['clib_order']
            del dirdict['clib_order']
        except KeyError:
            pass

    def get_extensions(self):
        if self.extensionObjs is None:
            self.extensionObjs = []
            for swigfile in self.dirdata['swigfiles']:
                # The [6:] in the following strips the "./SRC/" from
                # the beginning of the file names.  splitext(file)[0]
                # is the path to a file with the extension stripped.
                basename = os.path.splitext(swigfile)[0][6:]

                # swig 1.1 version
                modulename = os.path.splitext(basename + SWIGCFILEEXT)[0]
                sourcename = os.path.join(swigroot, basename+SWIGCFILEEXT)
                # swig 1.3 version                
##                modulename = '_' + basename
##                sourcename = os.path.join(swigroot, basename+'_wrap.cxx')
                
                extension = distutils.core.Extension(
                    name = os.path.join("oof2", "SWIG", modulename),
                    language = 'c++',
                    sources = [sourcename],
                    define_macros = platform['macros'],
                    extra_compile_args = platform['extra_compile_args'],
                    include_dirs = self.includeDirs + platform['incdirs'],
                    library_dirs = self.externalLibDirs + platform['libdirs'],
                    libraries = [self.libname] + self.externalLibs,
                                                        # + platform['libs'],
                    extra_link_args = platform['extra_link_args']
                    )

                self.extensionObjs.append(extension)
        return self.extensionObjs

    def get_shlib(self):
        if self.dirdata['cfiles']:
            return build_shlib.SharedLibrary(
                self.libname,
                sources=self.dirdata['cfiles'],
                extra_compile_args=platform['extra_compile_args'],
                include_dirs=self.includeDirs + platform['incdirs'],
                libraries=self.externalLibs,# + platform['libs'],
                library_dirs=self.externalLibDirs +
                platform['libdirs'],
                extra_link_args=platform['extra_link_args'])

    # Find all directories containing at least one swig input file.  These
    # are used to create the swigged python packages.  This is done by
    # traversing the DIR.py files, so that random leftover .swg files in
    # strange places don't create packages, and so that modules can be
    # included conditionally by HAVE_XXX tests in DIR.py files.
    def find_swig_pkgs(self):
        dict = {}
        for swigfile in self.dirdata['swigfiles']:
            dict[os.path.split(swigfile)[0]] = 1
        # pgks is a list of dirs containing swig files, relative to
        # the main OOF2 dir, eg, "./SRC/common"
        pkgs = dict.keys()
        # convert it to a list of dirs relative to the swigroot
        swigpkgs = []
        for pkg in pkgs:
            pkgdir = os.path.join(swigroot, pkg[6:]) # eg, SRC/SWIG/common
            pkgname = 'oof2.' + pkgdir[4:].replace('/', '.') # oof2.SWIG.common
            swigpkgs.append(pkgname)
            if not os.path.exists(pkgdir):
                os.makedirs(pkgdir)
        return swigpkgs

def moduleSort(moduleA, moduleB):
    if moduleA.ordering is not None:
        if moduleB.ordering is not None:
            return cmp(moduleA.ordering, moduleB.ordering)
        return -1
    else:                               # moduleA.ordering is None
        if moduleB.ordering is not None:
            return 1
        return cmp(moduleA.name, moduleB.name)

def allFiles(key):
    hierlist = [lib.dirdata[key] for lib in allCLibs.values()]
    flatlist = []
    for sublist in hierlist:
        flatlist.extend(sublist)
    return flatlist


def readDIRs(srcdir):
    dirfile = os.path.join(srcdir, DIRFILE)
    if os.path.exists(dirfile):
        # dirfile defines variables whose names are the same as the
        # ModuleInfo.dirdata keys.  The variables contain lists of
        # file names.
        localdict = {}
        execfile(dirfile, globals(), localdict)
        # Now the variables and functions defined in dirfile are in localdict.
        try:
            dirname = localdict['dirname']
            del localdict['dirname']
        except KeyError:
            pass
        
        try:
            clib = localdict['clib']
            del localdict['clib']
        except KeyError:
            pass
        else:
            clibinfo = getCLibInfo(clib)
            clibinfo.extractData(srcdir, localdict)
            
        try:
            pyfiles = localdict['pyfiles']
            del localdict['pyfiles']
        except KeyError:
            pass
        else:
            for filename in pyfiles:
                purepyfiles.append(os.path.join(srcdir, filename))
        
        # dirfile also contains a list of subdirectories to process.
        try:
            subdirs = localdict['subdirs']
            del localdict['subdirs']
        except KeyError:
            pass
        else:
            # At this point, all args in localdict should have been processed.
            if len(localdict) > 0:
                print "WARNING: unrecognized values", localdict.keys(), \
                      "in", dirfile
            for subdir in subdirs:
                readDIRs(os.path.join(srcdir, subdir))

##########

# Find all python packages and subpackages in a directory by looking
# for __init__.py files.

def find_pkgs():
    pkglist = []
    os.path.walk('SRC', _find_pkgs, pkglist)
    return pkglist

def _find_pkgs(pkglist, dirname, subdirs):
    if os.path.exists(os.path.join(dirname, '__init__.py')):
        pkglist.append(dirname)

##########

def swig_clibs(dry_run, force):
    srcdir = os.path.abspath('SRC')
    for clib in allCLibs.values():
        for swigfile in clib.dirdata['swigfiles']:
            # run_swig requires a src dir and an input file path
            # relative to it.  The '+1' in the following line strips
            # off a '/', so that sfile doesn't look like an absolute
            # path.
            sfile = os.path.abspath(swigfile)[len(srcdir)+1:]
            run_swig(srcdir='SRC', swigfile=sfile, destdir=swigroot,
                     cext=SWIGCFILEEXT,
                     include_dirs = ['SRC'],
                     dry_run=dry_run,
                     extra_args=platform['extra_swig_args'],
                     force=force
                     )

##########

# Get a file's modification time.  The time is returned as an integer.
# All we care about is that the integers are properly ordered.

def modification_time(file):
    return os.stat(file)[stat.ST_MTIME]

#########

# Define subclasses of the distutils build_ext and build_shlib class.
# We need subclasses so that oofconfig.h can be created before the
# files are compiled, and so that makedepend can be run.
# oof_build_xxxx contains the routines that are being added to both
# build_ext and build_shlib.

_dependencies_checked = 0
class oof_build_xxxx:
    def make_oofconfig(self):
        cfgfilename = os.path.normpath(os.path.join(self.build_temp,
                                                    'SRC/oofconfig.h'))
        self.distribution.data_files.append(('include', [cfgfilename]))
        # If oofconfig.h already exists, don't recreate it unless
        # forced to.  It would require everything that depends on it
        # to be recompiled unnecessarily.
        if self.force or not os.path.exists(cfgfilename):
            print "creating", cfgfilename
            if not self.dry_run:
                os.system('mkdir -p %s' % os.path.join(self.build_temp, 'SRC'))
                cfgfile = open(cfgfilename, "w")
                print >> cfgfile, """\
// This file was created automatically by the oof2 setup script.
// Do not edit it.
// Re-run setup.py to change the options.
#ifndef OOFCONFIG_H
#define OOFCONFIG_H
                """
                if HAVE_PETSC:
                    print >> cfgfile, '#define HAVE_PETSC 1'
                if HAVE_MPI:
                    print >> cfgfile, '#define HAVE_MPI 1'
                if DEVEL:
                    print >> cfgfile, '#define DEVEL ', DEVEL
                if self.check_header('<sstream>'):
                    print >> cfgfile, '#define HAVE_SSTREAM'
                else:
                    print >> cfgfile, '// #define HAVE_SSTREAM'
                print >> cfgfile, '#include <Python.h>'
                print >> cfgfile, "#endif"
                cfgfile.close()

    def check_header(self, headername):
        # Check that a c++ header file exists on the system.
        print "Testing for", headername
        # mktemp is insecure, but the better routines are only in
        # python2.3, and we don't otherwise require 2.3.
##        tmpfilename = tempfile.mktemp(suffix='.C')
##        tmpfile = open(tmpfilename, 'w')
        tmpfiled, tmpfilename = tempfile.mkstemp(suffix='.C')
        tmpfile = os.fdopen(tmpfiled, 'w')
        print >> tmpfile, """\
        #include %s
        int main(int, char**) { return 1; }
        """ % headername
        tmpfile.flush()
        try:
            try:
                ofiles = self.compiler.compile(
                    [tmpfilename],
                    extra_postargs=platform['extra_compile_args'] +
                                   platform['prelink_suppression_arg']
                    )
            except errors.CompileError:
                return 0
            ofile = ofiles[0]
            dir = os.path.split(ofile)[0]
            os.remove(ofiles[0])
            if dir:
                try:
                    os.removedirs(dir)
                except:
                    pass
            return 1
        finally:
            os.remove(tmpfilename)
    
    def find_dependencies(self, skip_swig):
        # distutils doesn't provide a makedepend-like facility, so we
        # have to do it ourselves.  This routine runs make_depend and
        # constructs a dictionary listing the dependencies of each .o
        # file and each swig-generated .C file.

        # TODO: This function is a mess.  Clean it up.
        
        # depdict[file] is a list of sources that the file depends
        # upon.
        depdict = {}

        # Run makedepend on the C++ files
        depfiled, depfilename = tempfile.mkstemp()
        cmd = '%(makedepend)s -f- -ISRC -I%(builddir)s -I%(buildsrc)s -- %(files)s > %(dependencies)s 2> /dev/null' \
              % {'makedepend' : MAKEDEPEND,
                 'files' : string.join(allFiles('cfiles'), ' '),
                 'dependencies' : depfilename,
                 'builddir' : self.build_temp,
                 'buildsrc' : os.path.join(self.build_temp, 'SRC')
                 }
##        print cmd
        print "running makedepend on .C files."
        os.system(cmd)
        # Read and parse the makedepend output.  Each nontrivial line
        # is of the form "target: source [source [...]]"
        depfile = os.fdopen(depfiled, 'r')
        for line in depfile.readlines():
            sline = line.strip()
            if sline and sline[0] != '#':
                target, sources = sline.split(': ')
                # The actual target that will be built is in the
                # build_temp directory.  makedepend assumes that it's
                # in the same directory as the source file, so we have
                # to modify the target name.
                realtarget = os.path.normpath(os.path.join(self.build_temp,
                                                           target))
                for source in sources.split():
                    try:
                        depdict[realtarget].append(source)
                    except KeyError:
                        depdict[realtarget] = [source]
        os.remove(depfilename)
        # makedepend omits the obvious dependency ("target.o:
        # target.C"), but we need it.
        for file in allFiles('cfiles'):
            ofile = os.path.normpath(
                os.path.join(self.build_temp, os.path.splitext(file)[0]+'.o'))
            try:
                depdict[ofile].append(file)
            except KeyError:
                depdict[ofile] = [file]

        # .C and.py Files in the SWIG directory depend on those in the
        # SRC directory.  Run makedepend on the swig source files.
        if not skip_swig:
            depfilename = tempfile.mktemp()
            cmd = '%(makedepend)s -f- -I. -ISRC -I%(builddir)s -- %(files)s > %(dependencies)s 2> /dev/null' \
                  % {'makedepend' : MAKEDEPEND,
                     'files' : string.join(allFiles('swigfiles'), ' '),
                     'dependencies' : depfilename,
                     'builddir' : self.build_temp
                     }
            print "running makedepend on .swg files."
##            print cmd
            os.system(cmd)
            depfile = open(depfilename, 'r')
            for line in depfile.readlines():
                sline = line.strip()
                if sline and sline[0] != '#':
                    target, sources = sline.split(': ')
                    targetbase = os.path.splitext(target)[0]
                    # Use [6:] to strip "./SRC/" from source paths
                    targetc = os.path.normpath(
                        os.path.join(swigroot, targetbase[6:] + SWIGCFILEEXT))
                    targetpy = os.path.normpath(
                        os.path.join(swigroot, targetbase[6:] + '.py'))
                    for source in sources.split():
                        try:
                            depdict[targetc].append(source)
                        except KeyError:
                            depdict[targetc] = [source]
                        try:
                            depdict[targetpy].append(source)
                        except KeyError:
                            depdict[targetpy] = [source]
            os.remove(depfilename)
            # Add in the implicit dependencies on the .swg files.
            for file in allFiles('swigfiles'):
                # file is ./SRC/dir/whatver.swg
                base = os.path.splitext(file)[0][6:] # dir/whatever
                cfile = os.path.normpath(os.path.join(swigroot,
                                                      base+SWIGCFILEEXT))
                pyfile = os.path.normpath(os.path.join(swigroot, base+'.py'))
                try:
                    depdict[cfile].append(file)
                except KeyError:
                    depdict[cfile] = [file]
                try:
                    depdict[pyfile].append(file)
                except:
                    depdict[pyfile] = [file]
            # Add in the implicit dependencies on the .spy files.
            for underpyfile in allFiles('swigpyfiles'):
                base = os.path.splitext(underpyfile)[0] # drop .spy
                pyfile = os.path.normpath(os.path.join(swigroot,base[6:]+'.py'))
                try:
                    depdict[pyfile].append(underpyfile)
                except KeyError:
                    depdict[pyfile] = [underpyfile]
        # end if not skip_swig
        return depdict

    # Remove out-of-date target files
    def clean_targets(self, depdict):
        outofdate = False
        if not self.dry_run:
            for target, sources in depdict.items():
                if os.path.exists(target):
                    targettime = modification_time(target)
                    sourcetime = max([modification_time(x) for x in sources])
                    if sourcetime > targettime:
##                        print 'removing', target
                        os.remove(target)
                        outofdate = True
                else:
                    outofdate = True
        if outofdate:
            ## TODO: Remove the .so file.  I can't figure out how to
            ## find its name at this point, though.
            pass

    def clean_dependencies(self, skip_swig):
        global _dependencies_checked
        if not _dependencies_checked:
            depdict = self.find_dependencies(skip_swig)
            self.clean_targets(depdict)
            _dependencies_checked = 1

class oof_build_ext(build_ext.build_ext, oof_build_xxxx):
    description = "build the python extension modules for OOF2"
    user_options = build_ext.build_ext.user_options + [
        ('skip-swig', None, "don't run swig")]
    boolean_options = ['skip-swig']
    def initialize_options(self):
        self.skip_swig = None           # must be None, not 0, or
                                        # set_undefined_options won't work
        build_ext.build_ext.initialize_options(self)
    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('skip_swig', 'skip_swig'))
        build_ext.build_ext.finalize_options(self)
    # build_extensions is called by build_ext.run().
    def build_extensions(self):
        self.compiler.add_include_dir(os.path.join(self.build_temp, 'SRC'))
        self.compiler.add_include_dir('SRC')
        self.compiler.add_library_dir(self.build_lib)

        if self.debug:
            self.compiler.define_macro('DEBUG')
            self.compiler.undefine_macro('NDEBUG')
        # Make the automatically generated .h files.
        self.make_oofconfig()
        # Run makedepend
        self.clean_dependencies(self.skip_swig)
        # Generate swigged .C and .py files
        if not self.skip_swig:
            swig_clibs(self.dry_run, self.force)
                                          
        # Build the swig extensions by calling the distutils base
        # class function
        build_ext.build_ext.build_extensions(self)

class oof_build_shlib(build_shlib.build_shlib, oof_build_xxxx):
    user_options = build_shlib.build_shlib.user_options + [
        ('skip-swig', None, "don't run swig"),
        ('blas-libraries=', None, "libraries for blas and lapack"),
        ('blas-link-args=', None, "link arguments required for blas and lapack")
        ]
    boolean_options = build_shlib.build_shlib.boolean_options + ['skip-swig']
    def initialize_options(self):
        self.skip_swig = None
        self.blas_libraries = None
        self.blas_link_args = None
        build_shlib.build_shlib.initialize_options(self)
    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('skip_swig', 'skip_swig'),
                                   ('blas_libraries', 'blas_libraries'),
                                   ('blas_link_args', 'blas_link_args'),
                                   ('libraries', 'libraries'),
                                   ('library_dirs', 'library_dirs'))
        build_shlib.build_shlib.finalize_options(self)
    def build_libraries(self, libraries):
        self.make_oofconfig()
        self.clean_dependencies(self.skip_swig)
        self.compiler.add_include_dir(os.path.join(self.build_temp, 'SRC'))
        self.compiler.add_include_dir('SRC')
        if self.debug:
            self.compiler.define_macro('DEBUG')

        # The blas libs and arguments weren't actually put into the
        # SharedLibrary objects when they were created, because we
        # didn't know until now whether or not the user had provided
        # alternates.  It's time to either use the predefined values
        # from "platform" or to use the command line arguments.
        for library in libraries:
            if self.blas_libraries is not None:
                library.libraries.extend(string.split(self.blas_libraries))
            else:
                library.libraries.extend(platform['blas_libs'])
            if self.blas_link_args is not None:
                library.extra_link_args.extend(
                    string.split(self.blas_link_args))
            else:
                library.extra_link_args.extend(platform['blas_link_args'])
        build_shlib.build_shlib.build_libraries(self, libraries)

# Add the 'skip-swig' argument to the base build command.
class oof_build(build.build):
    sep_by = " (separated by '%s')" % os.pathsep
    user_options = build.build.user_options + [
        ('skip-swig', None, "don't run swig"),
        ('libraries=', None, 'external libraries to link with'),
        ('library-dirs=', None,
         "directories to search for external libraries" + sep_by),
        ('blas-libraries=', None, "libraries for blas and lapack"),
        ('blas-link-args=', None, "link args for blas and lapack")]
    boolean_options = build.build.boolean_options + ['skip-swig']
    def initialize_options(self):
        self.skip_swig = 0
        self.libraries = None
        self.library_dirs = None
        self.blas_libraries = None
        self.blas_link_args = None
        build.build.initialize_options(self)
    
###################################################

## Modify the build_py command so that it creates oof2config.py.  The
## file is created in the build_lib directory so that it gets
## installed at the top level beside the oof2 package.

class oof_build_py(build_py.build_py):
    def run(self):
        self.make_oof2config()
        build_py.build_py.run(self)
    def make_oof2config(self):
        cfgscriptname = os.path.join(self.build_lib, 'oof2config.py')
        cfgscript = open(cfgscriptname, 'w')

        install = self.get_finalized_command('install')
        build_shlib = self.get_finalized_command('build_shlib')
        install_shlib = self.get_finalized_command('install_shlib')
      
        print >> cfgscript, 'root = "%s"' % os.path.abspath('.')
        print >> cfgscript, 'version = "%s"' % self.distribution.get_version()
        print >> cfgscript, 'prefix = "%s"' % install.prefix
        idirs = build_shlib.include_dirs + [
            os.path.abspath('SRC'),
            os.path.join(install.prefix, 'include')
            ] + platform['incdirs']
        print >> cfgscript, 'swig_include = ', [os.path.abspath('SRC')]
        print >> cfgscript, 'extra_compile_args =', \
              platform['extra_compile_args']
        print >> cfgscript, 'include_dirs =', idirs
        ldirs = [install_shlib.install_dir]
        print >> cfgscript, 'library_dirs =', ldirs
        librs = [lib.name for lib in install_shlib.shlibs]
        print >> cfgscript, 'libraries =', librs
        print >> cfgscript, 'extra_link_args =', platform['extra_link_args']
        print >> cfgscript, "import sys; sys.path.append(root)"
        cfgscript.close()


###################################################        

class oof_clean(clean.clean):
    user_options = clean.clean.user_options + [
        ('most', 'm', 'remove libraries and scripts, but not binary dist.'),
        ('swig', None, 'remove swig output files')]
    boolean_options = clean.clean.boolean_options + ['most', 'swig']

    def initialize_options(self):
        clean.clean.initialize_options(self)
        self.most = None
        self.swig = None
        
    def run(self):
        if self.most and not self.all:
            for d in [self.build_lib, self.build_scripts]:
                if os.path.exists(d):
                    remove_tree(d,dry_run=self.dry_run)
                else:
                    log.warn("'%s' does not exist -- can't clean it.", d)
        if self.swig and os.path.exists(swigroot):
            remove_tree(swigroot, dry_run=self.dry_run)
        clean.clean.run(self)
    
###################################################

# Make sure that there's an __init__.py file in each SWIG subdirectory
# and that the subdirectory is in the package list.  This is necessary
# because we've found the swig packages by searching for .swg files,
# but they might be in subdirectories of swig-less directories.  If
# the swig-less directories aren't listed as packages, then their
# __init__.py files won't be installed, and the swig directories can't
# be imported.

def _make_init_py(dummy, dirname, subdirs):
    initpy = os.path.join(dirname, '__init__.py')
    if not os.path.exists(initpy):
        file = open(initpy, 'w')
        print >> file, "# This file intentionally left blank."
        file.close()
    pkgname = 'oof2.' + dirname[4:].replace('/', '.') # remove 'SRC/'
    if pkgname not in pkg_list:
        pkg_list.append(pkgname)


###################################################

def set_dirs():
    global swigroot, exampledir, datadir, docdir
    swigroot = os.path.join('SRC', SWIGDIR)
    # Splitting and reassembling paths makes them portable to systems
    # that don't use '/' as the path separator.
    exampledir = os.path.join(*EXAMPLEDIR.split('/'))
    datadir = os.path.join(*DATADIR.split('/'))
    docdir = os.path.join(*DOCDIR.split('/'))

def get_global_args():
    # The --enable-xxxx flags in the command line have to be obtained
    # *before* we parse the DIR.py files, and the DIR.py files must be
    # parsed before calling distutils.core.setup (because the list of
    # source files comes from DIR.py).  But distutils.core.setup handles
    # the command line arguments, so we have to look for and remove the
    # --enable-xxxx flags here.

    # TODO: the more elegant way to do this would be to add a separate
    # distutils command that reads the DIR.py files and is always run
    # before any other command.  Then the --enable-xxxx flags could be
    # global distutils options, since they'd be processed *before* DIR.py
    # was read.

    global HAVE_MPI, HAVE_PETSC, DEVEL

    HAVE_MPI = _get_oof_arg('--enable-mpi')
    HAVE_PETSC = _get_oof_arg('--enable-petsc')
    DEVEL = _get_oof_arg('--enable-devel')

def _get_oof_arg(arg):
    # Search for an argument which begins like "arg" -- if found,
    # return the trailing portion if any, or 1 if none, and remove the
    # argument from sys.argv.
    for s in sys.argv:
        splits = s.split('=')
        if splits[0] == arg:
            sys.argv.remove(s)
            if len(splits) > 1:         # found an =
                return splits[1]
            return 1                    # just a plain arg
    return 0                            # didn't find arg
        
platform = {}

def set_platform_values():
    ## Set platform-specific flags that don't specifically depend on
    ## OOF2 stuff.  They're stored in a dictionary just to keep things
    ## tidy.
    platform['extra_compile_args'] = []
    platform['macros'] = []
    platform['blas_libs'] = []
    platform['blas_link_args'] = []
    platform['libdirs'] = []
    platform['incdirs'] = [get_config_var('INCLUDEPY')]
    platform['extra_link_args'] = []
    platform['prelink_suppression_arg'] = []
    platform['extra_swig_args'] = []

    if os.path.exists('/usr/local/lib'):
        platform['libdirs'].append('/usr/local/lib')
    if os.path.exists('/usr/site/lib'):
        platform['libdirs'].append('/usr/site/lib')
    if os.path.exists('/usr/site/include'):
        platform['incdirs'].append('/usr/site/include')

    # The prelink-suppression argument is used when running the compiler
    # to test for libraries.  Such builds need to be reasonably clean in
    # terms of not creating a lot of auxiliary files, and it's OK if
    # they're a bit slow.  Currently this is only set on SGIs, and it
    # prevents the creation of the "ii_files" subdirectory for the
    # library-check compilations.
    platform['prelink_suppression_arg'] = []

    if sys.platform == 'darwin':
        platform['blas_link_args'].extend(['-faltivec', '-framework', 'vecLib'])
    elif sys.platform == 'linux2':
        platform['blas_libs'].extend(['lapack', 'blas', 'm', 'g2c'])
    elif sys.platform[:4] == 'irix':
        platform['extra_compile_args'].append('-LANG:std')
        platform['extra_link_args'].append('-LANG:std')
        platform['prelink_suppression_arg'].append('-no_prelink')
        platform['blas_libs'].extend(['lapack', 'blas', 'ftn', 'm'])

    if HAVE_MPI:
        # mpi.h
        basedirs = ['/usr/local', '/usr', '/usr/lib', '/usr/site', '/sw']
        mpisubdirs = ['.', 'mpich', 'mpi']
        mpidirs = [os.path.join(bdir, 'include', mdir) for bdir in basedirs
                   for mdir in mpisubdirs]
        incdir = find_file('mpi.h', mpidirs)
        if not incdir:
            print "Warning! Can't locate mpi.h!"
        else:
            if incdir not in platform['incdirs']:
                platform['incdirs'].append(incdir)
        # mpi++.h
        basedirs = ['/usr/local', '/usr', '/usr/lib', '/usr/site', '/sw']
        mpisubdirs = ['.', 'mpich', 'mpi', 'mpich/mpi2c++', 'mpi/mpi2c++']
        mpidirs = [os.path.join(bdir, 'include', mdir) for bdir in basedirs
                   for mdir in mpisubdirs]
        incdir = find_file('mpi++.h', mpidirs)
        if not incdir:
            print "Warning! Can't locate mpi++.h!"
        else:
            if incdir not in platform['incdirs']:
                platform['incdirs'].append(incdir)

        platform['extra_swig_args'].append('-DHAVE_MPI')
        if sys.platform == 'darwin':
            # TODO: Need to know what Mac needs for mpi++
            # This may have to be changed or removed just as in MPI below
            platform['libs'].extend(['mpich', 'pmpich'])
        elif sys.platform == 'linux2':
            # libpmpich++.a
            basedirs = ['/usr/local', '/usr', '/usr/lib', '/usr/site', '/sw']
            mpisubdirs = ['.', 'mpich', 'mpi', 'mpich/lib', 'mpi/lib']
            mpidirs = [os.path.join(bdir, mdir) for bdir in basedirs
                       for mdir in mpisubdirs]
            libdir = find_file('libpmpich++.a', mpidirs)
            if not libdir:
                print "Warning! Can't locate libpmpich++.a!"
            else:
                if libdir not in platform['libdirs']:
                    platform['libdirs'].append(libdir)
            #platform['libs'].extend(['pmpich++', 'mpich'])
        elif sys.platform[:4] == 'irix':
            pass

#######################

# just_run_swig() is called by make_distutils_dist.

# TODO: The right way to do this is to get rid of make_distutils_dist
# and use a customized sdist command that runs swig.  We have to
# include the swig output in the source distribution so that users
# don't have to get the archaic version of swig that we're using.

def just_run_swig():
    set_dirs()
    get_global_args()
    set_platform_values()
    readDIRs('.')
    swig_clibs(dry_run=False, force=True)

#######################

if __name__ == '__main__':
    set_dirs()
    get_global_args()
    set_platform_values()

    readDIRs('.')                       # Gather data from the DIR.py files.

    # Get the data to build the C++ extension modules.
    clibraries = allCLibs.values()
    clibraries.sort(moduleSort)
    extensions = []
    shlibs = []
    for clib in clibraries:
        extensions.extend(clib.get_extensions())
        shlib = clib.get_shlib()
        if shlib is not None:
            shlibs.append(shlib)

    # Construct the list of pure python packages. If a subdirectory has an
    # __init__.py file, then all the .py files in the subdirectory will
    # form a package.

    # find non-swigged files
    pkg_list = ['oof2']
    pkgs = find_pkgs()                      # ['SRC', 'SRC/common', ...]
    for pkg in pkgs:
        if pkg != 'SRC':
            pkgname = 'oof2.' + pkg[4:].replace('/', '.')
            pkg_list.append(pkgname)

    # Ask each CLibInfo object for the swigged python modules it
    # creates.  This creates SWIG subdirectories if necessary.
    for clib in allCLibs.values():
        pkg_list.extend(clib.find_swig_pkgs())

    # Make sure each SWIG subdirectory has an __init__.py file, and
    # that the subdirectory is in the package list.
    os.path.walk(swigroot, _make_init_py, None)

    # Find example files:
    examples = [os.path.join('examples', file)
                for file in os.listdir('examples')] + \
                [os.path.join('examples/extension', file)
                 for file in os.listdir('examples/extension')]
    examples = [file for file in examples
                if os.path.isfile(file)] # exclude dirs
    

    # If this script is being used to create a frozen executable, the
    # Python path has to be set the same way it is during actual
    # execution.
    py2app_options = dict(argv_emulation=True)
    sys.path.append('SRC')                  # so that py2app can find imports
    try:
        import pygtk
        pygtk.require("2.0")
    except:
        pass

    setupargs = dict(
        name = "oof2",
        version = "2.0.0",
        description = "Analysis of material microstructures, from NIST.",
        author = 'The NIST OOF Team',
        author_email = 'oof_manager@ctcms.nist.gov',
        url = "http://www.ctcms.nist.gov/oof/oof2/",
        scripts = ["oof2"],
        cmdclass = {"build" : oof_build,
                    "build_ext" : oof_build_ext,
                    "build_py" : oof_build_py,
                    "build_shlib": oof_build_shlib,
                    "clean" : oof_clean},
        packages = pkg_list,
        package_dir = {'oof2':'SRC'},
        shlibs = shlibs,
        ext_modules = extensions,
        data_files = [(exampledir, examples)]
        )

##    if have_py2app:
##        setupargs['app'] = ['SRC/common/oof.py']
##        setupargs['options'] = {'py2app': py2app_options}

    distutils.core.setup(**setupargs)
